import six

from pycrunch.cubes import fetch_cube, count
from scrunch.datasets import Variable
from scrunch.expressions import parse_expr, process_expr
from cr.cube.crunch_cube import CrunchCube


def variable_to_url(variable, dataset):
    """
    Get a variable alias, instance or url and return
    the url
    """
    if isinstance(variable, Variable):
        return variable.url
    elif isinstance(variable, six.string_types):
        if variable in dataset.keys():
            return dataset[variable].url
    return variable


def crtabs(dataset, variables, weight=None, filter=None, **measures):
    """Return CrunchCube representation of crosstab.

    :param dataset: Dataset instance
    :param variables: List of variable urls, aliases or Variable instances
    :param weight: Scrunch variable instance, alias or url
    :param filter: Scrunch filter expression
    """
    variables = [variable_to_url(var, dataset) for var in variables]
    if weight is not None:
        weight = variable_to_url(weight, dataset)
    if filter is not None:
        filter = process_expr(parse_expr(filter), dataset.resource)

    return CrunchCube(fetch_cube(
        dataset.resource, variables, count=count(), weight=weight, filter=filter, **measures))



import mock
from unittest import TestCase

import pytest
from scrunch.datasets import Variable
from scrunch.mutable_dataset import MutableDataset
from scrunch.variables import responses_from_map
from scrunch.helpers import subvar_alias

from pycrunch.shoji import Entity


def _any(self, column):
    return {
        'function': 'any',
        'args': [{
            'variable': self
        }, {
            'column': column
        }]
    }


def mr_in(mr_self, mr_alias, groups, parent_subvariables):
    """
    Similar helper as in examples.py. Has to return an `any` function.
    """
    return _any(mr_self, [parent_subvariables[subvar_alias(mr_alias, sv)].id
                          for sv in groups])


dataset_url = 'http://test.crunch.io/api/datasets/123/'
var_url = 'http://test.crunch.io/api/datasets/123/variables/0001/'
subvar1_url = 'http://test.crunch.io/api/datasets/123/variables/0001/subvariables/00001/'
subvar2_url = 'http://test.crunch.io/api/datasets/123/variables/0001/subvariables/00002/'


CATEGORY_MAP = {
    1: [2, 3],
    2: 1
}
CATEGORY_NAMES = {
    1: 'China',
    2: 'Other'
}

RESPONSE_MAP = {
    1: [1, 2]
}
RESPONSE_NAMES = {
    1: 'online'
}

RECODES_PAYLOAD = {
    'element': 'shoji:entity',
    'body': {
        'alias': 'alias',
        'derivation': {
            'function': 'combine_categories',
            'args': [
                {'variable': var_url},
                {'value': [
                    {'combined_ids': [2, 3], 'missing': False, 'id': 1, 'name': 'China'},
                    {'combined_ids': [1], 'missing': False, 'id': 2, 'name': 'Other'}
                ]}
            ]
        },
        'name': 'name',
        'description': ''
    }
}

COMBINE_RESPONSES_PAYLOAD = {
    'element': 'shoji:entity',
    'body': {
        'name': 'name',
        'description': '',
        'alias': 'alias',
        'derivation': {
            'function': 'combine_responses',
            'args': [
                {'variable': var_url},
                {'value': [
                    {'alias': 'alias_1', 'combined_ids': [subvar1_url, subvar2_url], 'name': 'online'}
                ]}
            ]
        }
    }
}


class TestCombine(TestCase):

    def test_validate_range_expression(self):
        test_map = {
            1: range(1, 5)
        }
        test_cats = {
            1: "China"
        }
        ds_res_mock = mock.MagicMock()
        variable_mock = mock.MagicMock()
        subvar_mock = mock.MagicMock(entity_url=subvar1_url)
        # mock the call to entity, this will happen on Variable.resource
        variable_mock.entity.subvariables.by.return_value = {
            'parent_1': subvar_mock,
            'parent_2': subvar_mock,
            'parent_3': subvar_mock,
            'parent_4': subvar_mock,
        }
        parent_var = Variable(variable_mock, ds_res_mock)
        modified_map = responses_from_map(parent_var, test_map, test_cats, 'test', 'parent')
        # subvar_url * 4 because we used the same mock for all subvars
        assert modified_map[0]['combined_ids'] == [subvar1_url] * 4

    def test_validate_integer(self):
        test_map = {
            1: 1
        }
        test_cats = {
            1: "China"
        }

        ds_res_mock = mock.MagicMock()
        variable_mock = mock.MagicMock()
        subvar_mock = mock.MagicMock(entity_url=subvar1_url)
        # mock the call to entity, this will happen on Variable.resource
        variable_mock.entity.subvariables.by.return_value = {
            'parent_1': subvar_mock
        }
        parent_var = Variable(variable_mock, ds_res_mock)
        modified_map = responses_from_map(parent_var, test_map, test_cats, 'test', 'parent')
        # subvar_url * 4 because we used the same mock for all subvars
        assert modified_map[0]['combined_ids'] == [subvar1_url]

    def test_combine_categories_unknown_alias(self):
        resource = mock.MagicMock()
        resource.body = {'name': 'mocked_dataset'}
        entity_mock = mock.MagicMock(entity_url=var_url)
        resource.variables.by.return_value = {
            'test': entity_mock
        }
        resource.variables.index = {}  # Var not present
        ds = MutableDataset(resource)
        with pytest.raises(ValueError) as err:
            ds.combine_categorical('unknown', CATEGORY_MAP, CATEGORY_NAMES, name='name', alias='alias')

        assert 'Entity mocked_dataset has no (sub)variable with a name or alias unknown' in str(err.value)

    def test_combine_categories_from_alias(self):
        resource = mock.MagicMock()
        resource.body = {'name': 'mocked_dataset'}
        entity_mock = mock.MagicMock()
        entity_mock.entity.self = var_url
        resource.variables.by.return_value = {
            'test': entity_mock,
        }
        resource.variables.index = {}
        ds = MutableDataset(resource)
        with pytest.raises(ValueError) as err:
            ds.combine_categorical('test', CATEGORY_MAP, CATEGORY_NAMES, name='name', alias='alias')
        ds.resource.variables.create.assert_called_with(RECODES_PAYLOAD)
        assert 'Entity mocked_dataset has no (sub)variable' in str(err.value)

    def test_combine_categories_from_entity(self):
        resource = mock.MagicMock()
        resource.body = {'name': 'mocked_dataset'}
        entity_mock = mock.MagicMock()
        entity_mock.entity.self = var_url
        resource.variables.by.return_value = {
            'test': entity_mock
        }
        resource.variables.index = {}  # Var not present

        # mock a Tuple object
        tuple_mock = mock.MagicMock()
        tuple_mock.entity.self = var_url

        entity = Variable(tuple_mock, resource)
        ds = MutableDataset(resource)
        with pytest.raises(ValueError) as err:
            ds.combine_categorical(entity, CATEGORY_MAP, CATEGORY_NAMES, name='name', alias='alias')
        ds.resource.variables.create.assert_called_with(RECODES_PAYLOAD)
        assert 'Entity mocked_dataset has no (sub)variable' in str(err.value)

    def test_combine_responses_unknown_alias(self):
        resource = mock.MagicMock()
        resource.entity.self = dataset_url

        # mock subvariables
        subvar_mock = mock.MagicMock()
        subvar_mock.entity.self = subvar1_url
        subvar2_mock = mock.MagicMock()
        subvar2_mock.entity.self = subvar2_url

        # mock parent variable
        entity_mock = mock.MagicMock()
        entity_mock.entity.self = var_url

        # add dictionaries return to by functions
        entity_mock.entity.subvariables.by.return_value = {
            'test_1': subvar_mock,
            'test_x': subvar2_mock
        }

        resource.variables.by.return_value = {
            'test': entity_mock
        }

        ds = MutableDataset(resource)
        with pytest.raises(ValueError) as err:
            ds.combine_multiple_response('test', RESPONSE_MAP, RESPONSE_NAMES, name='name', alias='alias')

        assert 'Unknown subvariables for variable' in str(err.value)

    def test_combine_responses_by_alias(self):
        resource = mock.MagicMock()
        resource.body = {'name': 'mocked_dataset'}
        resource.entity.self = dataset_url
        resource.variables.index = {}  # Var not present
        # mock subvariables
        subvar_mock = mock.MagicMock(entity_url=subvar1_url)
        subvar2_mock = mock.MagicMock(entity_url=subvar2_url)

        # mock parent variable
        entity_mock = mock.MagicMock()
        entity_mock.entity.self = var_url

        # add dictionaries return to by functions
        entity_mock.entity.subvariables.by.return_value = {
            'test_1': subvar_mock,
            'test_2': subvar2_mock
        }

        resource.variables.by.return_value = {
            'test': entity_mock
        }

        # make the actual response call
        ds = MutableDataset(resource)
        with pytest.raises(ValueError) as err:
            ds.combine_multiple_response('test', RESPONSE_MAP, RESPONSE_NAMES, name='name', alias='alias')
        resource.variables.create.assert_called_with(COMBINE_RESPONSES_PAYLOAD)
        assert 'Entity mocked_dataset has no (sub)variable' in str(err.value)

    def test_combine_responses_by_entity(self):
        resource = mock.MagicMock()
        resource.body = {'name': 'mocked_dataset'}
        resource.entity.self = dataset_url
        resource.variables.index = {}  # Var not present

        # mock subvariables
        subvar_mock = mock.MagicMock(entity_url=subvar1_url)
        subvar_mock.entity.self = subvar1_url
        subvar2_mock = mock.MagicMock(entity_url=subvar2_url)

        # mock parent variable
        entity_mock = mock.MagicMock()
        # need to name the var to actually build subvar names
        entity_mock.alias = 'test'
        entity_mock.resource.self = var_url

        # add dictionaries return to by functions
        entity_mock.resource.subvariables.by.return_value = {
            'test_1': subvar_mock,
            'test_2': subvar2_mock
        }

        resource.variables.by.return_value = {
            'test': entity_mock
        }

        ds = MutableDataset(resource)

        with pytest.raises(ValueError) as err:
            ds.combine_multiple_response(entity_mock, RESPONSE_MAP, RESPONSE_NAMES, name='name', alias='alias')
        resource.variables.create.assert_called_with(COMBINE_RESPONSES_PAYLOAD)
        assert 'Entity mocked_dataset has no (sub)variable' in str(err.value)


class TestRecode(TestCase):

    @mock.patch('scrunch.mutable_dataset.get_mutable_dataset')
    def test_recode_categoricals(self, get_dataset_mock):
        categories = [
            {
                'missing': False,
                'name': 'Heterosexual',
                'numeric_value': 1,
                'id': 1
            },
            {
                'missing': False,
                'name': 'Gay or lesbian',
                'numeric_value': 2,
                'id': 2
            },
            {
                'missing': False,
                'name': 'Bisexual',
                'numeric_value': 3,
                'id': 3
            },
            {
                'missing': False,
                'name': 'Other',
                'numeric_value': 4,
                'id': 4
            },
            {
                'missing': False,
                'name': 'Prefer not to say',
                'numeric_value': 5,
                'id': 5
            },
            {
                'missing': True,
                'name': 'skipped',
                'numeric_value': None,
                'id': 8
            },
            {
                'missing': True,
                'name': 'not asked',
                'numeric_value': None,
                'id': 9
            }
        ]
        table_mock = mock.MagicMock(metadata={
            '00001': {
                'id': '00001',
                'alias': 'sexuality',
                'type': 'categorical',
                'categories': categories
            }
        })

        ds_res = mock.MagicMock()
        ds_res.self = dataset_url
        ds_res.follow.return_value = table_mock
        dataset = MutableDataset(ds_res)
        dataset.create_categorical([
            {'id': 1, 'name': 'Straight', 'case': 'sexuality.any([1])'},
            {'id': 2, 'name': 'LGBTQ+', 'case': 'sexuality.any([2, 3, 4, 5])'}
        ], name='Sexuality 2', alias='sexuality2', multiple=False)

        ds_res.variables.create.assert_called_with({
            'element': 'shoji:entity',
            'body': {
                'name': 'Sexuality 2',
                'alias': 'sexuality2',
                'description': '',
                'notes': '',
                'expr': {
                    'function': 'case',
                    'args': [{
                        'column': [1, 2, -1],
                        'type': {
                            'value': {
                                'class': 'categorical',
                                'categories': [
                                    {'missing': False, 'id': 1, 'name': 'Straight', 'numeric_value': None},
                                    {'missing': False, 'id': 2, 'name': 'LGBTQ+', 'numeric_value': None},
                                    {'numeric_value': None, 'missing': True, 'id': -1, 'name': 'No Data'}
                                ]
                            }
                        }
                    }, {
                        'function': 'any',
                        'args': [
                            {'variable': 'http://test.crunch.io/api/datasets/123/variables/00001/'},
                            {'value': [1]}
                        ]
                    }, {
                        'function': 'any',
                        'args': [
                            {'variable': 'http://test.crunch.io/api/datasets/123/variables/00001/'},
                            {'value': [2, 3, 4, 5]}
                        ]
                    }]
                }
            }
        })

    @mock.patch('scrunch.mutable_dataset.get_mutable_dataset')
    def test_recode_multiple_responses(self, get_dataset_mock):
        dataset_id = '123'
        categories = [
            {
                'numeric_value': 1,
                'selected': True,
                'id': 1,
                'name': 'selected',
                'missing': False
            },
            {
                'numeric_value': 2,
                'selected': False,
                'id': 2,
                'name': 'not selected',
                'missing': False
            },
            {
                'numeric_value': 9,
                'missing': True,
                'id': 9,
                'name': 'not asked'
            },
            {
                'numeric_value': 8,
                'missing': True,
                'id': 8,
                'name': 'skipped'
            }
        ]
        Entity(mock.MagicMock(), **{
            'element': 'shoji:entity',
            'self': 'http://test.crunch.io/api/datasets/%s/variables/0001/' % dataset_id,
            # needed in order to simulate a Tuple, now Variable is inited with Tuple
            'entity_url': 'http://test.crunch.io/api/datasets/%s/variables/0001/' % dataset_id,
            'body': {
                'name': 'Q1',
                'subreferences': [
                    {
                        'alias': 'Q1_1',
                        'is_subvar': True,
                        'name': 'One'
                    },
                    {
                        'alias': 'Q1_2',
                        'is_subvar': True,
                        'name': 'Two'
                    },
                    {
                        'alias': 'Q1_3',
                        'is_subvar': True,
                        'name': 'Three'
                    }
                ],
                'missing_reasons': {
                    'skipped': 8,
                    'not asked': 9
                },
                'alias': 'Q1',
                'subvariables': [
                    'http://test.crunch.io/api/datasets/%s/variables/0001/subvariables/000a/' % dataset_id,
                    'http://test.crunch.io/api/datasets/%s/variables/0001/subvariables/000b/' % dataset_id,
                    'http://test.crunch.io/api/datasets/%s/variables/0001/subvariables/000c/' % dataset_id
                ],
                'dataset_id': dataset_id,
                'type': 'multiple_response',
                'id': '0001',
                'categories': categories,
                'description': 'Multiple Response Example',
                'notes': '',
            }

        })
        table_mock = mock.MagicMock(metadata={
            '00001': {
                'id': '00001',
                'alias': 'sexuality',
                'type': 'categorical',
                'notes': '',
                'categories': categories
            }
        })
        ds_res = mock.MagicMock()
        ds_res.self = dataset_url
        ds_res.follow.return_value = table_mock
        dataset = MutableDataset(ds_res)
        subvar_mock = mock.MagicMock()
        subvar_mock.self = var_url
        subvar_mock.id = 'subvar'
        subvariables = {
            'Q1_1': subvar_mock,
            'Q1_2': subvar_mock,
            'Q1_3': subvar_mock,
        }

        dataset.create_categorical([
            {'id': 1, 'name': 'Q1_recoded_1', 'case': mr_in(var_url, 'Q1', [1, 2], subvariables)},
            {'id': 2, 'name': 'Q1_recoded_2', 'case': mr_in(var_url, 'Q1', [3], subvariables)}
        ], alias='Q1_recoded', name='Q1_recoded', multiple=True)

        # Test how the recoded var was created.
        ds_res.variables.create.assert_called_with({
            'element': 'shoji:entity',
            'body': {
                'name': 'Q1_recoded',
                'description': '',
                'notes': '',
                'alias': 'Q1_recoded',
                'derivation': {
                    'function': 'array',
                    'args': [{
                        'function': 'select',
                        'args': [{
                            'map': {
                                '0001': {
                                    'function': 'case',
                                    'args': [{
                                        'column': [1, 2],
                                        'type': {
                                            'value': {
                                                'class': 'categorical',
                                                'categories': [
                                                    {'selected': True, 'numeric_value': None, 'missing': False, 'id': 1, 'name': 'Selected'},
                                                    {'selected': False, 'numeric_value': None, 'missing': False, 'id': 2, 'name': 'Not selected'}
                                                ]
                                            }
                                        }
                                    }, {
                                        'function': 'any',
                                        'args': [
                                            {'variable': 'http://test.crunch.io/api/datasets/123/variables/0001/'},
                                            {'column': ['subvar', 'subvar']}
                                        ]
                                    }],
                                    'references': {
                                        'alias': 'Q1_recoded_1',
                                        'name': 'Q1_recoded_1'
                                    }
                                },
                                '0002': {
                                    'function': 'case',
                                    'args': [{
                                        'column': [1, 2],
                                        'type': {
                                            'value': {
                                                'class': 'categorical',
                                                'categories': [
                                                    {'selected': True, 'numeric_value': None, 'missing': False, 'id': 1, 'name': 'Selected'},
                                                    {'selected': False, 'numeric_value': None, 'missing': False, 'id': 2, 'name': 'Not selected'}
                                                ]
                                            }
                                        }
                                    }, {
                                        'function': 'any',
                                        'args': [
                                            {'variable': 'http://test.crunch.io/api/datasets/123/variables/0001/'},
                                            {'column': ['subvar']}
                                        ]
                                    }],
                                    'references': {
                                        'alias': 'Q1_recoded_2',
                                        'name': 'Q1_recoded_2'
                                    }
                                }
                            }
                        }, {
                            'value': [
                                '0001',
                                '0002'
                            ]
                        }]
                    }]
                }
            }
        })



import requests
import six

if six.PY2:  # pragma: no cover
    from urlparse import urljoin
else:
    from urllib.parse import urljoin


# ID's for default categories
SELECTED_ID = 1
NOT_SELECTED_ID = 2
NO_DATA_ID = -1


DEFAULT_MULTIPLE_RESONSE_CATEGORIES = [
    {'id': SELECTED_ID, 'name': 'Selected', 'missing': False, 'numeric_value': None, 'selected': True},
    {'id': NOT_SELECTED_ID, 'name': 'Not selected', 'missing': False, 'numeric_value': None, 'selected': False},
]


class ReadOnly(object):
    """
    class for protecting undesired writes to attributes
    """
    def __init__(self, resource):
        # need to call parent to make sure we call other mixin's __init__
        object.__setattr__(self, "resource", resource)
        object.__setattr__(self, "url", resource.self)

    def __setattr__(self, attr, value):
        if attr in self._IMMUTABLE_ATTRIBUTES:
            raise AttributeError(
                "Can't edit attibute '%s'" % attr)
        if attr in self._MUTABLE_ATTRIBUTES:
            raise AttributeError('use the edit() method for '
                                 'mutating attributes')
        object.__setattr__(self, attr, value)


def is_relative_url(url):
    return url.startswith(('.', '/'))


def abs_url(expr, base_url):
    """
    Converts an expression that may contain relative references to variable
     URLs into absolute URLs.

    This is necessary when using the derivation expression from a variable
    entity endpoint and sending it back to the variable catalog endpoint.
    """
    if isinstance(expr, dict):
        for k in expr:
            if k == 'variable':
                if is_relative_url(expr[k]):
                    expr[k] = urljoin(base_url, expr[k])
            elif isinstance(expr[k], dict):
                expr[k] = abs_url(expr[k], base_url)
            elif isinstance(expr[k], list):
                expr[k] = [abs_url(xpr, base_url) for xpr in expr[k]]
    elif isinstance(expr, list):
        expr = [abs_url(xpr, base_url) for xpr in expr]
    return expr


def subvar_alias(parent_alias, response_id):
    return '%s_%d' % (parent_alias, response_id)


def download_file(url, filename):
    if url.startswith('file://'):
        # Result is in local filesystem (for local development mostly)
        import shutil
        shutil.copyfile(url.split('file://', 1)[1], filename)
    else:
        r = requests.get(url, stream=True)
        with open(filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:   # filter out keep-alive new chunks
                    f.write(chunk)
    return filename


def get_else_case(case, responses):
    """
    When creating a categorical like:
        responses=[
            {'id': 1, 'name': '30', 'case': 'age == 30'}
            {'id': 2, 'name': '40', 'case': 'age == 40'}
            {'id': 3, 'name': 'Other', 'case': 'else'}
        ]
    build the case for else: 'not (age==30) and not (age==40)'
    """
    if case == 'else':
        case = ' and '.join(
            ['not({})'.format(_case['case']) for _case in responses if _case['case'] != 'else']
        )
    return case


def else_case_not_selected(case, responses, missing_case):
    """
    When creating a multiple response through derive_multiple_response with:
        categories = [
            {
                'id': 1,
                'name': 'Coke',
                'case': 'q1 in [1]',
                'missing_case': 'missing(q1)'
            },
            {
                'id': 2,
                'name': 'Pepsi',
                'case': 'q1 in [2]',
                'missing_case': 'missing(q1)'
            },
            {
                'id': 3,
                'name': 'Other',
                'case': 'else',
                'missing_case': 'missing(screener4)'
            }]
    In this case, the else case needs to be manually built in the form:
        - not selected: '((q1 in [1]) or (q1 in [2]) and not (missing(screener4)))'
    """
    if case == 'else' and missing_case:
        missing = ' or '.join(
            ['({})'.format(_case['case']) for _case in responses if _case['case'] != 'else']
        )
        missing = '({})'.format(missing)
        missing += ' and not ({}) '.format(missing_case)
        return missing
    return None


def validate_categories(categories):
    """
    Categories can be defined as simple as:
        categories=[
            {'id': 1, 'name': 'Yes', 'selected': True},
            {'id': 2, 'name': 'No'},
            {'id': 3, 'name': 'Maybe'},
            {'id': 4, 'name': 'Missing', 'missing': True}
        ]
    This method takes care of validating that only 1 is selected
    and to fill the definition above to match the API requirement:
        categories=[
            {'id': 1, 'name': 'Yes', 'missing': False, 'numeric_value': None, 'selected': True},
            {'id': 2, 'name': 'No', 'missing': False, 'numeric_value': None, 'selected': False},
            {'id': 3, 'name': 'Maybe', 'missing': False, 'numeric_value': None, 'selected': False},
            {'id': 4, 'name': 'Missing', 'missing': True, 'numeric_value': None, 'selected': False}
        ]
    """
    defaults = {'missing': False, 'numeric_value': None, 'selected': False}
    selected_count = 0
    for category in categories:
        if category.get('selected'):
            selected_count += 1
        if not category.get('id'):
            raise ValueError('An "id" must be provided to all categories')
        if not category.get('name'):
            raise ValueError('A "name" must be provided to all categories')
    if selected_count > 1 or selected_count == 0:
        raise ValueError('Categories must define one category as selected')
    _categories = []
    for category in categories:
        default = defaults.copy()
        default.update(category)
        _categories.append(default)
    return _categories


def case_expr(rules, name, alias, categories=DEFAULT_MULTIPLE_RESONSE_CATEGORIES):
    """
    Given a set of rules, return a `case` function expression to create a
    variable.
    """
    expression = {
        'references': {
            'name': name,
            'alias': alias,
        },
        'function': 'case',
        'args': [{
            'column': [category['id'] for category in categories],
            'type': {
                'value': {
                    'class': 'categorical',
                    'categories': categories
                }
            }
        }]
    }
    for rule in rules:
        expression['args'].append(rule)
    return expression


def _validate_category_rules(categories, rules):
    """
    validate that rules and categories are of same size
    """

    if not ((len(categories) - 1) <= len(rules) <= len(categories)):
        raise ValueError(
            'Amount of rules should match categories (or categories -1)'
        )


def shoji_entity_wrapper(body, **kwargs):
    """
    receives a dictionary and wraps its content on a body keyed dictionary
    with the appropiate shoji:<class> attribute
    """
    payload = {
        'element': 'shoji:entity',
        'body': body
    }
    payload.update(**kwargs)
    return payload


def shoji_catalog_wrapper(index, **kwargs):
    """
    receives a dictionary and wraps its content on a body keyed dictionary
    with the appropiate shoji:<class> attribute
    """
    payload = {
        'element': 'shoji:catalog',
        'index': index
    }
    payload.update(**kwargs)
    return payload



# coding: utf-8

"""
This example shows a more complex example using different recodes and combines
to create new variables.
"""

import os
import tempfile
from unittest import TestCase

from fixtures import NEWS_DATASET, NEWS_DATASET_ROWS, mr_in, RECODES_CSV_OUTPUT
from scrunch import connect
from scrunch.streaming_dataset import get_streaming_dataset

HOST = os.environ['SCRUNCH_HOST']
username = os.environ['SCRUNCH_USER']
password = os.environ['SCRUNCH_PASS']


site = connect(username, password, HOST)


class TestRecodes(TestCase):
    def test_recodes(self):
        # Create a dataset for usage
        ds = site.datasets.create({
            'element': 'shoji:entity',
            'body': {
                'name': 'test_recodes',
                'table': {
                    'element': 'crunch:table',
                    'metadata': NEWS_DATASET
                },
                'streaming': 'streaming'
            }
        }).refresh()
        dataset = get_streaming_dataset(ds.body.id, site)
        print("Dataset %s created" % dataset.id)

        # Add data rows
        total = dataset.stream_rows(NEWS_DATASET_ROWS)
        dataset.push_rows(total)

        # Recode a new single response variable
        agerange = dataset.create_categorical([
            {'id': 1, 'name': 'Underage', 'case': 'age < 18'},
            {'id': 2, 'name': 'Millennials', 'case': 'age > 18 and age < 25'},
            {'id': 3, 'name': 'Gen X', 'case': 'age < 35 and age >= 25'},
            {'id': 4, 'name': 'Grown ups', 'case': 'age < 60 and age >= 35'},
            {'id': 5, 'name': '60+', 'case': 'age >= 60'}
        ], alias='agerange', name='Age range', multiple=False)

        print("Variable %s created" % agerange.alias)

        # Recode a new multiple response variable from an existing multiple
        # response variable
        origintype = dataset.create_categorical([
            {'id': 1, 'name': "Online",
             # Mixed support for using "category"(subvariables really) IDs
             'case': mr_in(dataset, 'newssource', [1, 2, 3, 4])},  # Only in the helper
            {'id': 2, 'name': "Print", 'case': mr_in(dataset, 'newssource', [5, 6])},
            {'id': 3, 'name': "Tv", 'case': mr_in(dataset, 'newssource', [7, 9])},
            {'id': 4, 'name': "Radio", 'case': mr_in(dataset, 'newssource', [8, 10])},
        ], alias='origintype', name="News source by type", multiple=True)

        print("Variable %s created" % origintype.alias)

        # Add an exclusion filter
        dataset.exclude('agerange == 1')  # Remove underage

        # MAYBE: Create a new numeric expression based on arbitrary rules

        # Copy a variable
        origintype_copy = dataset.copy_variable(
            origintype,
            name='Copy of origintype',
            alias='origintype_copy')
        print("Variable %s created" % origintype_copy.alias)

        onlinenewssource = dataset.combine_categories(
            origintype_copy,
            map={
                1: 1,
                2: [2, 3, 4]
            },
            categories={
                1: "online",
                2: "notonline",
            },
            alias="onlinenewssource",
            name='Online or not')
        print('Created combination: %s' % onlinenewssource.alias)

        # Combine a single categorical - Combine with subvar 3 on the wrong place
        over35 = dataset.combine_categories(
            agerange,
            map={
                1: [1, 2, 3],
                2: [4, 5]
            },
            categories={
                1: 'under35',
                2: 'over35'
            },
            name='over 35?',
            alias='over35')
        print('Created combination: %s' % over35.alias)

        self.maxDiff = None
        # Export some rows
        output = tempfile.NamedTemporaryFile('rw', delete=True)
        dataset.export(output.name)
        result = [l.strip() for l in output.read().strip().split('\n')]
        expected = RECODES_CSV_OUTPUT.split('\n')
        # Rows are unordered under streaming conditions
        self.assertEqual(sorted(result), sorted(expected))
        output.close()
        ds.delete()



import pytest
from mock import patch, MagicMock
from unittest import TestCase

from pycrunch.cubes import fetch_cube, count
from scrunch.streaming_dataset import StreamingDataset
from scrunch.tests.test_datasets import TestDatasetBase
from scrunch.cubes import crtabs, variable_to_url
from scrunch.datasets import Variable


class TestCubes(TestDatasetBase, TestCase):

    @patch('scrunch.cubes.fetch_cube')
    def test_crtabs_passes_string_arguments(self, mock_fetch_cube):
        """
        Test url aliases are converted to urls
        """
        ds_mock = self._dataset_mock()
        ds = StreamingDataset(ds_mock)
        variables = ['var1_alias', 'var2_alias']
        urls = [variable_to_url(var, ds) for var in variables]
        crtabs(dataset=ds, variables=variables)
        mock_fetch_cube.assert_called_once_with(
            ds.resource, urls, count=count(), weight=None, filter=None)

    @patch('scrunch.cubes.fetch_cube')
    def test_weight_to_url(self, mock_fetch_cube):
        """
        Test weight alias is converted to url
        """
        ds_mock = self._dataset_mock()
        ds = StreamingDataset(ds_mock)
        variables = ['var1_alias', 'var2_alias']
        weight_url = variable_to_url('var3_alias', ds)
        urls = [variable_to_url(var, ds) for var in variables]
        crtabs(dataset=ds, variables=variables, weight='var3_alias')
        mock_fetch_cube.assert_called_once_with(
            ds.resource, urls, count=count(), weight=weight_url, filter=None)

    @patch('scrunch.cubes.fetch_cube')
    def test_pass_filter_expression(self, mock_fetch_cube):
        ds_mock = self._dataset_mock()
        ds = StreamingDataset(ds_mock)
        variables = ['var1_alias', 'var2_alias']
        urls = [variable_to_url(var, ds) for var in variables]
        crtabs(dataset=ds, variables=variables, filter='var1_alias > 1')
        processed_filter = {
            'function': '>',
            'args': [
                {'variable': 'https://test.crunch.io/api/datasets/123456/variables/0001/'},
                {'value': 1}
            ]
        }
        mock_fetch_cube.assert_called_once_with(
            ds.resource, urls, count=count(), filter=processed_filter, weight=None)


"""
This module provides basic support for converting expression strings (defined
in a python-like DSL) into crunch expression objects.

For example, the expression 'disposition == 0 or exit_status == 0' would
be transformed by this module's parser into:

        {
            'function': 'or',
            'args': [
                {
                    'function': '==',
                    'args': [
                        {
                            'variable': 'disposition'
                        },
                        {
                            'value': 0
                        }
                    ]
                },
                {
                    'function': '==',
                    'args': [
                        {
                            'variable': 'exit_status'
                        },
                        {
                            'value': 0
                        }
                    ]
                }
            ]
        }

Its important to note that the expression objects produced by this module's
parser are not ready for being sent to crunch, as they refer to variables
by `alias` rather than by `variable_id` (which is a URL). So, this module
also provides a `process_expr` function that creates an expression object
ready for the crunch API.
"""

import ast
import copy

import six

import scrunch
from scrunch.variables import validate_variable_url


if six.PY2:
    from urllib import urlencode
else:
    from urllib.parse import urlencode

ARRAY_TYPES = ('categorical_array', 'multiple_response')

CRUNCH_FUNC_MAP = {
    'valid': 'is_valid',
    'missing': 'is_missing',
    'bin': 'bin',
    'selected': 'selected',
    'not_selected': 'not_selected',
}

CRUNCH_METHOD_MAP = {
    'any': 'any',
    'all': 'all',
    'duplicates': 'duplicates',
    'bin': 'bin',
    'selected': 'selected',
    'not_selected': 'not_selected',
}

# according to http://docs.crunch.io/#function-terms
BINARY_FUNC_OPERATORS = [
    '+',
    '-',
    '*',
    '/',
    '//',
    '^',
    '%',
    '&',
    '|',
    '~',
]

COMPARISSON_OPERATORS = [
    '==',
    '!=',
    '=><=',
    '<',
    '>',
    '<=',
    '>=',
    '~=',
    'in',
    'and',
    'or',
    'not',
]

COMPARISSON_FUNCS = [
    # 'between',
    'all',
    'any',
]

BUILTIN_FUNCTIONS = []

NOT_IN = object()


def _nest(args, func, concatenator=None):
    """
    :param args: list of arguments that need nesting
    :param func: the function that applies to tuple of args
    :param concatenator: the concatenator or tuples, usually 'or' 'and'
    :return:
    """
    if not concatenator:
        # in case of is_missing and is_valid we need to concatenate
        # multiple arguments and nest them
        concatenator = func
    # for the moment we are just nesting and & or
    if func not in ['or', 'and', 'is_missing', 'is_valid'] or len(args) < 3:
        return {
            'function': concatenator,
            'args': args
        }
    return {
        'function': concatenator,
        'args': [args[0], _nest(args[1:], func, concatenator)]
    }


def unfold_list(_list):
    if any(isinstance(x, ast.Call) for x in _list):
        new_list = list()
        for e in _list:
            if isinstance(e, ast.Call):
                name = e.func.id
                if name == 'r':
                    try:
                        lower = e.args[0].n
                        upper = e.args[1].n
                        r_list = r(lower, upper)
                        for elem in r_list:
                            new_list.append(ast.Num(elem))
                    except Exception:
                        raise AttributeError(
                            "function 'r' needs 2 integer arguments")
                else:
                    return _list
            else:
                new_list.append(e)
        return new_list
    else:
        return _list


def r(lower, upper):
    return list(range(lower, upper + 1))


def parse_expr(expr):

    def _parse(node, parent=None):
        obj = {}
        args = []
        op = None
        func_type = None

        if isinstance(node, ast.AST):
            # Get the current node fields.
            fields = list(ast.iter_fields(node))

            # "Terminal" nodes. Recursion ends with these guys.
            if isinstance(node, ast.Name):
                _id = fields[0][1]

                # A function identifier.
                if getattr(node, '_func_type', None) == 'function':
                    return _id

                # A variable identifier.
                return {
                    'variable': _id
                }
            elif isinstance(node, ast.Num) or isinstance(node, ast.Str):
                if isinstance(parent, ast.Call) \
                        and 'func' in parent._fields:
                    _id = fields[0][1]
                    return {
                        'variable': _id
                    }

                _val = fields[0][1]
                return {
                    'value': _val
                }
            elif isinstance(node, ast.Add):
                return '+'
            elif isinstance(node, ast.Sub):
                return '-'
            elif isinstance(node, ast.Mult):
                return '*'
            elif isinstance(node, ast.Div):
                return '/'
            elif isinstance(node, ast.FloorDiv):
                return '//'
            elif isinstance(node, ast.Pow):
                return '^'
            elif isinstance(node, ast.Mod):
                return '%'
            elif isinstance(node, ast.BitAnd):
                return '&'
            elif isinstance(node, ast.BitOr):
                return '|'
            elif isinstance(node, ast.Invert):
                return '~'
            elif isinstance(node, ast.Eq):
                return '=='
            elif isinstance(node, ast.NotEq):
                return '!='
            elif isinstance(node, ast.Lt):
                return '<'
            elif isinstance(node, ast.LtE):
                return '<='
            elif isinstance(node, ast.Gt):
                return '>'
            elif isinstance(node, ast.GtE):
                return '>='
            elif isinstance(node, ast.In):
                return 'in'
            elif isinstance(node, ast.NotIn):
                return NOT_IN
            elif isinstance(node, ast.List) or isinstance(node, ast.Tuple):
                _list = fields[0][1]
                # checks for special helper functions like `r`
                _list = unfold_list(_list)
                if not (all(isinstance(el, ast.Str) for el in _list) or
                        all(isinstance(el, ast.Num) for el in _list)):
                    # Only list-of-int or list-of-str are currently supported
                    raise ValueError('Only list-of-int or list-of-str are currently supported')

                return {
                    'value': [
                        getattr(el, 's', None) or getattr(el, 'n')
                        for el in _list
                    ]
                }
            elif isinstance(node, ast.Attribute) \
                    and isinstance(parent, ast.Call):
                # The variable.
                _id_node = fields[0][1]
                if not isinstance(_id_node, ast.Name):
                    msg = (
                        'calling methods of "{}" object not allowed, '
                        'variable name expected.'
                    ).format(type(_id_node).__name__)
                    raise SyntaxError(msg)

                _id = _parse(_id_node, parent=node)

                # The 'method'.
                method = fields[1][1]
                if method not in CRUNCH_METHOD_MAP.keys():
                    raise ValueError(
                        'unknown method "{}", valid methods are: [{}]'.format(
                            method,
                            ', '.join(CRUNCH_METHOD_MAP.keys())
                        )
                    )

                return _id, CRUNCH_METHOD_MAP[method]

            # "Non-terminal" nodes.
            else:
                for _name, _val in fields:
                    if not isinstance(node, ast.UnaryOp) and (
                            isinstance(_val, (ast.BoolOp, ast.UnaryOp, ast.Compare, ast.Call))):
                        # Descend.
                        obj.update(_parse(_val, parent=node))
                    elif isinstance(_val, ast.And):
                        op = 'and'
                    elif isinstance(_val, ast.Or):
                        op = 'or'
                    elif isinstance(_val, ast.Not):
                        op = 'not'
                    elif isinstance(_val, ast.Mult):
                        op = '*'
                    elif isinstance(_val, ast.Add):
                        op = '+'
                    elif isinstance(_val, ast.Div):
                        op = '/'
                    elif isinstance(_val, ast.Sub):
                        op = '-'
                    elif _name == 'left':
                        left = _parse(_val, parent=node)
                        args.append(left)
                    elif _name == 'right':
                        right = _parse(_val, parent=node)
                        args.append(right)
                    elif _name == 'func' and isinstance(_val, ast.Attribute):
                        # Method-like call. Example:
                        #       variable.any([1,2])
                        func_type = 'method'
                        setattr(_val, '_func_type', func_type)
                        left, op = _parse(_val, parent=node)
                        args.append(left)
                    elif _name == 'func' and isinstance(_val, ast.Name):
                        # Function call. Example:
                        #       valid(birthyear, birthmonth)
                        func_type = 'function'
                        setattr(_val, '_func_type', func_type)
                        _id = _parse(_val, parent=node)
                        if _id not in CRUNCH_FUNC_MAP.keys():
                            raise ValueError(
                                'unknown method "{}", valid methods are: [{}]'.format(
                                    _id,
                                    ', '.join(CRUNCH_METHOD_MAP.keys())
                                )
                            )
                        op = CRUNCH_FUNC_MAP[_id]
                    elif _name == 'ops':
                        if len(_val) != 1:
                            raise ValueError('only one logical operator at a time')
                        op = _parse(_val[0], parent=node)
                    elif _name == 'comparators' or _name == 'args':  # right
                        if len(_val) == 0:
                            continue

                        if func_type == 'method':
                            if len(_val) > 1:
                                raise ValueError('1 argument expected, got {}'.format(len(_val)))

                            if op == 'duplicates':
                                # No parameters allowed for 'duplicates'.
                                raise ValueError('No parameters allowed for "duplicates"')

                        for arg in _val:
                            right = _parse(arg, parent=node)

                            # For method calls, we only allow list-of-int
                            # parameters.
                            if _name == 'args' and func_type == 'method':
                                if not isinstance(right.get('value'), list):
                                    raise ValueError(
                                        'expected list, got "{}"'.format(
                                            type(right.get('value'))
                                        )
                                    )

                            args.append(right)

                    elif _name in ('keywords', 'starargs', 'kwargs') and _val:
                        # We don't support these in function/method calls.
                        raise ValueError('unsupported call with argument "{}"'.format(_name))
                    elif _name == 'operand' and isinstance(node, ast.UnaryOp):
                        right = _parse(_val, parent=node)
                        args.append(right)
                    elif isinstance(_val, list):
                        for arg in _val:
                            args.append(_parse(arg, parent=node))
                    elif isinstance(_val, ast.BinOp):
                        op = _parse(_val.op, _val)
                        args.append(_parse(_val.left, _val))
                        args.append(_parse(_val.right, _val))

                if op:
                    if op is NOT_IN:
                        # Special treatment for the `not in` operator.
                        obj = {
                            'function': 'not',
                            'args': [
                                {
                                    'function': 'in',
                                    'args': []
                                }
                            ]
                        }
                    elif op in CRUNCH_FUNC_MAP.values() \
                            and isinstance(args, list) and len(args) > 1:
                        obj = {
                            'function': 'or',
                            'args': []
                        }
                    else:
                        obj = {
                            'function': op,
                            'args': []
                        }

                if args and 'args' in obj:
                    if op is NOT_IN:
                        # Special treatment for the args in a `not in` expr.
                        obj['args'][0]['args'] = args
                    elif op in CRUNCH_FUNC_MAP.values() \
                            and isinstance(args, list) and len(args) > 1:
                        for arg in args:
                            obj['args'].append(
                                {
                                    'function': op,
                                    'args': [arg]
                                }
                            )
                        # concatenate with or when there is more than
                        # 2 arguments in the list
                        obj = _nest(obj['args'], op, concatenator='or')
                    else:
                        obj = _nest(args, op)

            # special case when te expr is just a variable
            # make sure we parse it's content
            if not obj and isinstance(node, ast.Expression):
                return _parse(node.body)

        return obj

    if expr is None:
        return dict()

    return _parse(ast.parse(expr, mode='eval'))


def get_dataset_variables(ds):
    table = ds.follow("table", urlencode({
        'limit': 0
    }))

    # Build the variables dict, using `alias` as the key.
    variables = dict()
    for _id, var in table.metadata.items():
        var['id'] = _id
        variables[var['alias']] = var

        if var['type'] in ARRAY_TYPES:
            subreferences = var.get('subreferences', {})
            for subvar_id, subvar in subreferences.items():
                subvar['is_subvar'] = True
                subvar['id'] = subvar_id
                subvar['parent_id'] = _id
                subvar['type'] = 'categorical'
                subvar['description'] = ''
                subvar['categories'] = copy.deepcopy(var['categories'])
                variables[subvar['alias']] = subvar

    return variables


def process_expr(obj, ds):
    """
    Given a Crunch expression object (or objects) and a Dataset entity object
    (i.e. a Shoji entity), this function returns a tuple, the first element is
    new expression object (or a list of new expression objects) with all
    variable aliases transformed into variable URLs and the second element
    of the tuple is a flag indicating if the expressions needs nesting/wrapping
    in `or` functions (for the case when an array variable is passed).
    """

    base_url = ds.self
    variables = get_dataset_variables(ds)
    var_index = ds.variables.index

    def get_subvariables_resource(var_url):
        variable = var_index[var_url].entity
        subvariables = variable.subvariables['index']
        return [(sv['id'], sv['alias']) for sv in subvariables.values()]

    def adapt_multiple_response(var_url, values):
        """
        Converts multiple response arguments
        to column.
        :return: tuple of the new args for multiple_response and
        a flag to indicate we don't need recursive nesting of this
        expression.
        """
        # convert value --> column and change ids to aliases
        aliases = get_subvariables_resource(var_url)

        # Some derived variables will append a # to the subvariables id's
        # so we need to strip those charactes out
        cat_to_ids = [
            tup[0] for tup in aliases if int(tup[1].strip('#').split('_')[-1]) in values]
        return [{'variable': var_url}, {'column': cat_to_ids}], False

    def ensure_category_ids(subitems, variables=variables):
        var_id = None
        _subitems = []

        def variable_id(variable_url):
            return variable_url.split('/')[-2]

        def category_ids(var_id, var_value, variables=variables):
            value = None
            if isinstance(var_value, list) or isinstance(var_value, tuple):
                # {'values': [val1, val2, ...]}
                value = []
                for val in var_value:
                    if str(val).isdigit():
                        # val1 is an id already
                        value.append(val)
                        continue
                    for var in variables:
                        if variables[var]['id'] == var_id:
                            if 'categories' in variables[var]:
                                for cat in variables[var]['categories']:
                                    if cat['name'] == val:
                                        value.append(cat['id'])
                            else:
                                # variable has no categories, return original
                                # list of values
                                value = var_value

            elif isinstance(var_value, str):
                for var in variables:
                    # if the variable is a date, don't try to process it's categories
                    if variables[var]['type'] == 'datetime':
                        return var_value
                    if variables[var]['id'] == var_id and 'categories' in variables[var]:
                        found = False
                        for cat in variables[var]['categories']:
                            if cat['name'] == var_value:
                                value = cat['id']
                                found = True
                                break
                        if not found:
                            raise ValueError("Couldn't find a category id for category %s in filter for variable %s" % (var_value, var))
                    elif 'categories' not in variables[var]:
                        return var_value

            else:
                return var_value
            return value

        # special case for multiple_response variables
        if len(subitems) == 2:
            if 'value' in subitems[1] and 'variable' in subitems[0]:
                var_url = subitems[0]['variable']
                if var_url in var_index and var_index[var_url]['type'] == 'multiple_response':
                    return adapt_multiple_response(var_url, subitems[1]['value'])

        for item in subitems:
            if isinstance(item, dict) and 'variable' in item:
                var_id = variable_id(item['variable'])
            elif isinstance(item, dict) and 'value' in item:
                item['value'] = category_ids(var_id, item['value'])
            _subitems.append(item)

        return _subitems, True

    def _process(obj, variables):
        op = None
        arrays = []
        values = []
        subvariables = []
        needs_wrap = True

        # inspect function, then inspect variable, if multiple_response,
        # then change in --> any
        if 'function' in obj and 'args' in obj:
            if obj['function'] == 'in':
                args = obj['args']
                if 'variable' in args[0]:
                    try:
                        if variables.get(args[0]['variable'])['type'] == 'multiple_response':
                            obj['function'] = 'any'
                    except TypeError:
                        raise ValueError("Invalid variable alias '%s'" % args[0]['variable'])

        for key, val in obj.items():
            if isinstance(val, dict):
                obj[key] = _process(val, variables)
            elif isinstance(val, list) or isinstance(val, tuple):
                subitems = []
                for subitem in val:
                    if isinstance(subitem, dict):
                        subitem = _process(subitem, variables)
                        if 'subvariables' in subitem:
                            arrays.append(subitem.pop('subvariables'))
                        elif 'value' in subitem:
                            values.append(subitem)
                    subitems.append(subitem)

                has_value = any('value' in item for item in subitems
                    if not str(item).isdigit())

                has_variable = any('variable' in item for item in subitems
                    if not str(item).isdigit())
                if has_value and has_variable:
                    subitems, needs_wrap = ensure_category_ids(subitems)
                obj[key] = subitems
            elif key == 'variable':
                var = variables.get(val)
                if var:
                    # TODO: We shouldn't stitch URLs together, use the API
                    if var.get('is_subvar'):
                        obj[key] = '%svariables/%s/subvariables/%s/' \
                                   % (base_url, var['parent_id'], var['id'])
                    else:
                        obj[key] = '%svariables/%s/' % (base_url, var['id'])

                    if var['type'] in ARRAY_TYPES:
                        subvariables = [
                            '%svariables/%s/subvariables/%s/'
                            % (base_url, var['id'], subvar_id)
                            for subvar_id in var.get('subvariables', [])
                        ]
                else:
                    raise ValueError("Invalid variable alias '%s'" % val)

            elif key == 'function':
                op = val

        if subvariables:
            obj['subvariables'] = subvariables

        if arrays and op in ('any', 'all', 'is_valid', 'is_missing') and needs_wrap:
            # Support for array variables.

            if len(arrays) != 1:
                raise ValueError

            real_op = 'in'
            expansion_op = 'or'
            if op == 'all':
                real_op = '=='
                expansion_op = 'and'
            elif op == 'is_valid':
                real_op = 'all_valid'
            elif op == 'is_missing':
                real_op = 'all_missing'

            if op in ('is_valid', 'is_missing'):
                if len(values) != 0:
                    raise ValueError

                # Just swap the op. Yep, that's it.
                obj['function'] = real_op
            else:
                if len(values) != 1:
                    raise ValueError

                subvariables = arrays[0]
                value = values[0]

                if op == 'all':
                    if len(value['value']) != 1:
                        raise ValueError
                    value['value'] = value['value'][0]

                if len(subvariables) == 1:
                    obj['function'] = real_op
                    obj['args'][0] = {'variable': subvariables[0]}
                    obj['args'][1] = value
                else:
                    obj = {
                        'function': expansion_op,
                        'args': []
                    }
                    args_ref = obj['args']
                    for i, subvar in enumerate(subvariables):
                        args_ref.append(
                            {
                                'function': real_op,
                                'args': [
                                    {'variable': subvar},
                                    value
                                ]
                            }
                        )
                        if i < len(subvariables) - 2:
                            args_ref.append(
                                {'function': expansion_op, 'args': []}
                            )
                            args_ref = args_ref[-1]['args']

        return obj

    if isinstance(obj, list):
        return [
            _process(copy.deepcopy(element), variables) for element in obj
        ]
    else:
        return _process(copy.deepcopy(obj), variables)


def prettify(expr, ds=None):
    """
    Translate the crunch expression dictionary to the string representation.

    :param expr: crunch expression
    :param ds: dataset instance
    :return: string representation of the expression
    """
    assert isinstance(expr, dict), "Dictionary is expected"

    operators = BINARY_FUNC_OPERATORS + COMPARISSON_OPERATORS
    methods = {m[1]: m[0] for m in CRUNCH_METHOD_MAP.items()}
    functions = {f[1]: f[0] for f in CRUNCH_FUNC_MAP.items()}

    def _resolve_variable(var):
        is_url = validate_variable_url(var)

        if not is_url:
            return var
        elif not isinstance(ds, scrunch.datasets.BaseDataset):
            raise Exception(
                'Valid Dataset instance is required to resolve variable urls '
                'in the expression'
            )
        return ds.resource.session.get(var).payload.body.alias

    def _resolve_variables(_expr):
        new_expr = dict(
            function=_expr['function'],
            args=[]
        )
        for arg in _expr['args']:
            if 'function' in arg:
                # arg is a function, resolve inner variables
                new_expr['args'].append(_resolve_variables(arg))
            elif 'variable' in arg:
                # arg is a variable, resolve
                new_expr['args'].append(
                    {'variable': _resolve_variable(arg['variable'])}
                )
            else:
                # arg is neither a variable or function, pass as is
                new_expr['args'].append(arg)
        return new_expr

    def _transform(f, args, nest=False):
        result = ''
        if f in operators:
            if len(args) == 1:
                result = '%s %s' % (f, args[0])
            else:
                op = ' %s ' % f
                result = op.join(str(x) for x in args)
        elif f in methods:
            if f in ['selected', 'not_selected']:
                result = '%s(%s)%s' % (methods[f], args[0], ', '.join(str(x) for x in args[1:]))
            else:    
                result = '%s.%s(%s)' % (
                    args[0], methods[f], ', '.join(str(x) for x in args[1:])
                )
        elif f in functions:
            result = '%s(%s)' % (functions[f], args[0])
        else:
            raise Exception('Unknown function "%s"' % f)

        if nest:
            result = '(%s)' % result

        return result

    def _quote_value(v):
        # escape the quotes from the string, also escape the backslash
        return "'{}'".format(
            v.replace("\\", "\\\\").replace("\'", "\\\'")
        )

    def _process(fragment, parent=None):
        _func = fragment.get('function')

        if _func is None:
            # This is not a function, but a plain argument

            if 'value' in fragment:
                # This argument is a value, not a variable
                value = fragment['value']

                if isinstance(value, six.string_types):
                    # Must escape single-quote from string value
                    value = _quote_value(value)

                return value

            return list(fragment.values())[0]

        args = [_process(arg, _func) for arg in fragment['args']]
        child_functions = [
            arg.get('function')
            for arg in fragment['args'] if arg.get('function') is not None
        ]
        has_child_and_or = 'or' in child_functions
        nest = parent is not None and (
            has_child_and_or or
            (parent == 'or' and len(child_functions) > 1) or
            _func == 'or'
        )
        return _transform(_func, args, nest=nest)

    return _process(_resolve_variables(expr))



import collections
import json
import re

import six

import pycrunch
import scrunch.datasets
from scrunch.exceptions import (InvalidPathError, InvalidReferenceError,
                                OrderUpdateError)

NAME_REGEX = re.compile(r'^\|$|^\|?([\w\s,&\(\)\-\/\\]+\|?)+$', re.UNICODE)


class Path(object):
    def __init__(self, path):
        if not isinstance(path, six.string_types):
            raise TypeError('The path must be a string object')

        if six.PY2:
            regex_match = re.match(NAME_REGEX, path.decode('utf-8'))
        else:
            regex_match = re.match(NAME_REGEX, path)

        if not regex_match:
            raise InvalidPathError(
                'Invalid path %s: it contains invalid characters.' % path
            )

        self.path = path

    @property
    def is_root(self):
        return self.path == '|'

    @property
    def is_absolute(self):
        return self.path.startswith('|')

    @property
    def is_relative(self):
        return not self.is_absolute

    def get_parts(self):
        return [part.strip() for part in self.path.split('|') if part]

    def __str__(self):
        return self.path

    def __repr__(self):
        return self.__str__()


class Group(object):

    INDENT_SIZE = 4

    def __init__(self, obj, order, parent=None):
        self.name = list(obj.keys())[0]
        self.order = order
        self.parent = parent
        self.elements = collections.OrderedDict()

        # Load all the elements.
        for element in obj[self.name]:
            if isinstance(element, six.string_types):
                _id = element.split('/')[-2]
                # NOTE: instantiating Variable/Dataset here seems overkill to
                # me. While its as simple as `Dataset(dataset.entity)` for the
                # `dataset` tuple below, for the Variable we would first need
                # to instantiate it's `Dataset`, going through all this order
                # machinery again...
                if 'datasets' not in element or 'variables' in element:
                    # 1. relative variable URL: ../<id>/
                    # 2. compl variable URL: /api/datasets/<id>/variables/<id>/
                    var = self.order.vars.get(_id)
                    if var:
                        self.elements[var.alias] = var
                elif 'datasets' in element and 'variables' not in element:
                    # 3. it's a dataset URL
                    dataset = self.order.datasets.get(_id)
                    if dataset:
                        self.elements[dataset.id] = dataset
                    else:
                        refresh_ds = self.order.catalog.refresh()['index']
                        for key, obj in refresh_ds.items():
                            if _id in key:
                                self.elements[obj.id] = obj
                                break
            elif isinstance(element, dict):
                subgroup = Group(element, order=self.order, parent=self)
                self.elements[subgroup.name] = subgroup
            # TODO unreached code
            elif isinstance(element, scrunch.datasets.Variable):
                self.elements[element.alias] = element
            elif isinstance(element, scrunch.datasets.BaseDataset):
                self.elements[element.id] = element
            else:
                raise TypeError('Invalid OrderObject %s' % element)

    def __str__(self):
        def _get_elements(group):
            elements = []
            for key, obj in list(group.elements.items()):
                if isinstance(obj, Group):
                    elements.append({key: _get_elements(obj)})
                # TODO unreached code
                elif isinstance(obj, (scrunch.datasets.Variable,
                                      scrunch.datasets.BaseDataset)):
                    elements.append(obj.name)
                else:
                    elements.append(obj.name)
            return elements

        str_elements = _get_elements(self)
        return json.dumps(str_elements, indent=self.INDENT_SIZE)

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        return self.itervalues()

    def itervalues(self):
        for item in self.elements.values():
            yield item

    def iterkeys(self):
        for key in self.elements.keys():
            yield key

    def keys(self):
        return list(self.iterkeys())

    def values(self):
        return list(self.itervalues())

    def items(self):
        return zip(self.iterkeys(), self.itervalues())

    def __getitem__(self, path):
        if not isinstance(path, six.string_types):
            raise TypeError('arg 1 must be a string')

        path = Path(path)

        if path.is_root and self.is_root:
            return self

        if path.is_absolute and not self.is_root:
            raise InvalidPathError(
                'Absolute paths can only be used on the root Group.'
            )

        group = self
        for part in path.get_parts():
            try:
                group = group.elements[part]
            except KeyError:
                raise InvalidPathError(
                    'Invalid path %s: element %s does not exist.' % (path, part)
                )
            except AttributeError:
                raise InvalidPathError(
                    'Invalid path %s: element %s is not a Group.' % (path, part)
                )
            if not isinstance(group, Group):
                raise InvalidPathError(
                    'Invalid path %s: element %s is not a Group.' % (path, part)
                )

        return group

    def __contains__(self, item):
        return item in self.elements and isinstance(self.elements[item], Group)

    def __delitem__(self, item):
        if len(self.elements[item].keys()) == 0:
            del self.elements[item]
            self.order.update()
        else:
            raise ValueError(
                "Cannot delete Groups containing items. Group must be empty first"
            )

    @property
    def is_root(self):
        return self.parent is None and self.name == '__root__'

    @staticmethod
    def _validate_alias_arg(alias):
        if isinstance(alias, six.string_types):
            alias = [alias]
        if not isinstance(alias, collections.Iterable):
            raise ValueError(
                'Invalid list of aliases/ids/groups to be inserted'
                ' into the Group.'
            )
        if not all(isinstance(a, six.string_types) for a in alias):
            raise ValueError(
                'Only string references to aliases/ids/group names'
                ' are allowed.'
            )
        return alias

    def _validate_name_arg(self, name):
        if not isinstance(name, six.string_types):
            raise ValueError(
                'The name argument must be a string object.'
            )
        if '|' in name:
            raise ValueError(
                'The pipe (|) character is not allowed.'
            )
        if name in self.elements:
            raise ValueError(
                'A variable/sub-group named \'%s\' already exists.' % name
            )

        if six.PY2:
            regex_match = re.match(NAME_REGEX, name.decode('utf-8'))
        else:
            regex_match = re.match(NAME_REGEX, name)

        if not regex_match:
            raise ValueError("Invalid character in name: %s" % name)

        return name

    def _validate_reference_arg(self, reference):
        if not isinstance(reference, six.string_types):
            raise TypeError('Invalid reference. It must be a string.')
        if reference not in self.elements:
            raise InvalidReferenceError(
                'Invalid reference %s: it is not part of the current Group.'
                % reference
            )
        return reference

    def find(self, name):

        def _find(group):
            for _name, obj in group.elements.items():
                if isinstance(obj, Group):
                    result = _find(obj)
                    if result:
                        return result
                elif _name == name:
                    return group

        return _find(self)

    def find_group(self, name):

        def _find(group):
            if group.name == name:
                return group
            for _name, obj in group.elements.items():
                if isinstance(obj, Group):
                    result = _find(obj)
                    if result:
                        return result

        return _find(self)

    def insert(self, alias, position=0, before=None, after=None):
        elements = self._validate_alias_arg(alias)

        if not isinstance(position, int):
            raise ValueError('Invalid position. It must be an integer.')
        if position < -1 or position > len(self.elements):
            raise IndexError('Invalid position %d' % position)
        if position == 0 and (before or after):
            reference = self._validate_reference_arg(before or after)
            i = 0
            for name in self.elements.keys():
                if name in elements:
                    continue
                if name == reference:
                    if before:
                        position = i
                    elif after:
                        position = i + 1
                    break
                i += 1
        if position == -1:
            position = len(self.elements)

        elements_to_move = collections.OrderedDict()
        for element_name in elements:
            if element_name in self.elements:
                elements_to_move[element_name] = \
                    (self.elements[element_name], '__move__')
            else:
                current_group = self.order.group.find(element_name)
                if current_group:
                    # A variable.
                    elements_to_move[element_name] = \
                        (current_group, '__migrate_element__')
                else:
                    # Not a variable. A group, maybe?
                    group_to_move = self.order.group.find_group(element_name)
                    if group_to_move:
                        elements_to_move[element_name] = \
                            (group_to_move, '__migrate_group__')
                    else:
                        raise ValueError(
                            'Invalid alias/id/group name \'%s\'' % element_name
                        )

        # Make all necessary changes to the order structure.
        _elements = collections.OrderedDict()
        _non_targeted = list()

        for name in list(self.elements.keys()):
            if name in elements_to_move:
                continue
            _non_targeted.append((name, self.elements[name]))
        _non_targeted = list(reversed(_non_targeted))  # FIFO for .pop()

        total = len(_non_targeted) + len(elements_to_move)
        i = 0
        while i <= total:
            if i == position:
                for element_name in elements_to_move.keys():
                    obj, operation = elements_to_move[element_name]

                    if operation == '__move__':
                        _elements[element_name] = obj
                    elif operation == '__migrate_element__':
                        current_group = obj
                        element = current_group.elements[element_name]
                        del current_group.elements[element_name]
                        _elements[element_name] = element
                    elif operation == '__migrate_group__':
                        group_to_move = obj
                        orig_parent = group_to_move.parent
                        group_to_move.parent = self
                        del orig_parent.elements[element_name]
                        _elements[element_name] = group_to_move
            elif len(_non_targeted):
                element_name, obj = _non_targeted.pop()
                _elements[element_name] = obj

            i += 1

        self.elements = _elements

        # Update!
        self.order.update()

    def append(self, alias):
        self.insert(alias, position=-1)

    def reorder(self, items):
        existing_items = [name for name in self.elements.keys()]
        if len(items) != len(existing_items) or \
                not all(i in existing_items for i in items):
            raise ValueError('Invalid list of items for the reorder operation.')

        if items == existing_items:
            # Nothing to do.
            return

        _elements = collections.OrderedDict()
        for item in items:
            _elements[item] = self.elements[item]
        self.elements = _elements

        self.order.update()

    def create_group(self, name, alias=None, position=-1, before=None,
                     after=None):
        name = self._validate_name_arg(name)
        # when we want to create an empty group
        if not alias:
            self.elements[name] = Group(
                {name: []}, order=self.order, parent=self)
            self.order.update()
            return
        elements = self._validate_alias_arg(alias)
        position = 0 if (before or after) else position

        # create the new Group obj and insert all `elements`
        new_group = Group({name: []}, order=self.order, parent=self)
        new_group.insert(elements)

        # add the new Group to self.elements so that `insert` detects it
        self.elements[name] = new_group
        self.insert(new_group.name, position=position,
                    before=before, after=after)

    def rename(self, name):
        name = self._validate_name_arg(name)

        if self.is_root:
            raise ValueError(
                'Renaming the root Group is not allowed.'
            )

        if name in self.parent.elements:
            raise ValueError(
                'Parent Group \'%s\' already contains an element named \'%s\'.'
                % (self.parent.name, name)
            )

        if name == self.name:
            # Nothing to do.
            return

        # Rename!
        _elements = collections.OrderedDict()
        for current_name, obj in self.parent.elements.items():
            if current_name == self.name:
                _elements[name] = obj
            else:
                _elements[current_name] = obj
        self.name = name
        self.parent.elements = _elements

        # Update!
        self.order.update()

    def move(self, path, position=-1, before=None, after=None):
        position = 0 if (before or after) else position
        path = Path(path)
        if not path.is_absolute:
            raise InvalidPathError(
                'Invalid path %s: only absolute paths are allowed.' % path
            )
        target_group = self.order[str(path)]

        if target_group == self:
            raise InvalidPathError(
                'Invalid path %s: cannot move Group into itself.' % path
            )
        target_group.insert(
            self.name, position=position, before=before, after=after
        )


class Order(object):
    """
    How do you expect to use this class?

    examples:
        $ Project.order
        ...
        $ Project.order['|'].create_group(
            'FirstGroup',entities=[ds1.id, ds2.id], before..., after...)
        $ Project.order['|'].create_group('2ndGroup')
    """

    def __init__(self, catalog, order):
        self.catalog = catalog
        self.order = order
        self._load(refresh=False)

    def _load(self, refresh=True):
        """
        Do not call this method manually, this is intended for
        methods only
        """
        if refresh:
            self.catalog.refresh()
            self.order.refresh()
        self.group = Group({'__root__': self.order.graph}, order=self)

    def place(self, entity, path, position=-1, before=None, after=None):
        """
        place an entity into a specific place in the order hierarchy
        """
        position = 0 if (before or after) else position
        path = Path(path)
        if not path.is_absolute:
            raise InvalidPathError(
                'Invalid path %s: only absolute paths are allowed.' % path
            )
        target_group = self.group[str(path)]
        if isinstance(entity, scrunch.datasets.Variable):
            element = entity.alias
        elif isinstance(entity, scrunch.datasets.BaseDataset):
            element = entity.id
        else:
            raise TypeError('entity must be a `Variable` or `Dataset`')

        target_group.insert(
            element, position=position,
            before=before, after=after)

    def _prepare_shoji_graph(self):
        """
        returns shoji:graph ready for the server
        """

        def _get(group):
            _elements = []
            for obj in group.elements.values():
                if isinstance(obj, Group):
                    _elements.append({
                        obj.name: _get(obj)
                    })
                else:
                    url = obj.entity.self
                    _elements.append(url)
            return _elements

        return _get(self.group)

    def update(self):
        updated_order = {
            'element': 'shoji:order',
            'graph': self._prepare_shoji_graph()
        }
        try:
            # NOTE: Order has no Attribute edit
            self.order.put(updated_order)
        except pycrunch.ClientError as e:
            # Our update to the Hierarchical Order failed. Better reload.
            self._load(refresh=True)
            raise OrderUpdateError(str(e))

    # Proxy methods for the __root__ Group

    def __str__(self):
        return str(self.group)

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        return self.group.itervalues()

    def itervalues(self):
        return self.group.itervalues()

    def iterkeys(self):
        return self.group.iterkeys()

    def keys(self):
        return self.group.keys()

    def values(self):
        return self.group.values()

    def items(self):
        return self.group.items()

    def __getitem__(self, item):
        return self.group[item]


class DatasetVariablesOrder(Order):

    def _load(self, refresh=True):
        self.vars = self.catalog.by('id')
        super(DatasetVariablesOrder, self)._load(refresh=refresh)


class ProjectDatasetsOrder(Order):

    def _load(self, refresh=False):
        self.datasets = self.catalog.by('id')
        super(ProjectDatasetsOrder, self)._load(refresh=refresh)



# coding: utf-8

import json
from mock import Mock
from requests import Response
from unittest import TestCase

from pycrunch.shoji import Entity, Catalog, Order

from scrunch.order import InvalidPathError
from scrunch.datasets import Project, ProjectDatasetsOrder, get_personal_project

from .mock_session import MockSession


class TestProjectNesting(TestCase):
    def test_detect_correct_handler(self):
        session = Mock(
            feature_flags={'old_projects_order': True}
        )
        dataset_order = Order(session, **{
            'graph': []
        })
        datasets_catalog = Catalog(session, **{
            'index': {},
            'order': dataset_order
        })
        shoji_resource = Entity(session, **{
            'self': '/project/url/',
            'body': {},
            'index': {},
            'datasets': datasets_catalog
        })
        project = Project(shoji_resource)
        self.assertTrue(isinstance(project.order, ProjectDatasetsOrder))

        session = Mock(
            feature_flags={'old_projects_order': False}
        )
        shoji_resource = Entity(session, **{
            'self': '/project/url/',
            'body': {},
            'index': {},
            'graph': []
        })
        project = Project(shoji_resource)
        self.assertTrue(isinstance(project.order, Project))

    def test_create_subproject(self):
        session = MockSession()
        session.feature_flags = {'old_projects_order': False}
        shoji_resource = Entity(session, **{
            'self': 'http://example.com/project/url/',
            'body': {},
            'index': {},
            'graph': [],
        })

        # Setup the POST request and the fixture for the GET that happens after
        # the .refresh()
        created_project_url = 'http://example.com/project/2/'
        response = Response()
        response.status_code = 201
        response.headers = {
            'Location': created_project_url
        }
        session.add_post_response(response)
        session.add_fixture(created_project_url, {
            'self': created_project_url,
            'body': {},
            'index': {},
            'graph': [],
        })
        project = Project(shoji_resource)

        # Create a new project
        pa = project.order.create_project("Project A")
        self.assertTrue(isinstance(pa, Project))

        # Check that we sent the correct payload to the server
        self.assertEqual(pa.url, created_project_url)
        post_request = session.requests[-2]
        refresh_request = session.requests[-1]
        self.assertEqual(refresh_request.method, 'GET')
        self.assertEqual(post_request.method, 'POST')
        self.assertEqual(post_request.url, project.url)
        self.assertEqual(json.loads(post_request.body), {
            'element': 'shoji:entity',
            'body': {
                'name': 'Project A'
            }
        })

    def make_tree(self):
        session = MockSession()
        session.feature_flags = {'old_projects_order': False}

        #       A
        #     /   \
        #    B     C
        #    |
        #    D
        projects_res_url = 'http://example.com/api/projects/'
        a_res_url = 'http://example.com/api/projects/A/'
        b_res_url = 'http://example.com/api/projects/B/'
        c_res_url = 'http://example.com/api/projects/C/'
        d_res_url = 'http://example.com/api/projects/D/'
        a_payload = {
            'element': 'shoji:entity',
            'self': a_res_url,
            'catalogs': {
                'project': 'http://example.com/api/projects/'
            },
            'body': {
                'name': 'project A'
            },
            'index': {
                b_res_url: {
                    'id': 'idB',
                    'name': 'project B',
                    'icon': None,
                    'description': '',
                    'type': 'project'
                },
                c_res_url: {
                    'id': 'idC',
                    'name': 'project C',
                    'icon': None,
                    'description': '',
                    'type': 'project'
                }
            },
            'graph': [c_res_url, b_res_url]
        }
        b_payload = {
            'element': 'shoji:entity',
            'self': b_res_url,
            'catalogs': {
                'project': a_res_url
            },
            'body': {'name': 'project B'},
            'index': {
                d_res_url: {
                    'id': 'idD',
                    'name': 'project D',
                    'icon': None,
                    'description': '',
                    'type': 'project'
                }
            },
            'graph': [d_res_url]
        }
        c_payload = {
            'element': 'shoji:entity',
            'self': c_res_url,
            'catalogs': {
                'project': a_res_url
            },
            'body': {'name': 'project C'},
            'index': {},
            'graph': []
        }
        d_payload = {
            'element': 'shoji:entity',
            'self': d_res_url,
            'catalogs': {
                'project': b_res_url
            },
            'body': {'name': 'project D'},
            'index': {},
            'graph': []
        }
        projects_catalog = {
            'element': 'shoji:catalog',
            'self': projects_res_url,
            'index': {
                a_res_url: {
                    'name': 'project A'
                }
            }
        }
        session.add_fixture(a_res_url, a_payload)
        session.add_fixture(b_res_url, b_payload)
        session.add_fixture(c_res_url, c_payload)
        session.add_fixture(d_res_url, d_payload)
        session.add_fixture(projects_res_url, projects_catalog)
        return session

    def test_follow_path(self):
        a_res_url = 'http://example.com/api/projects/A/'
        d_res_url = 'http://example.com/api/projects/D/'

        session = self.make_tree()
        a_res = session.get(a_res_url).payload
        project_a = Project(a_res)
        project_c = project_a.order['| project C ']
        project_d = project_a.order['| project B | project D']
        self.assertTrue(isinstance(project_d, Project))
        self.assertEqual(project_d.resource.self, d_res_url)

        with self.assertRaises(InvalidPathError):
            project_a.order['| project B | Invalid']

    def test_rename(self):
        a_res_url = 'http://example.com/api/projects/A/'
        session = self.make_tree()
        a_res = session.get(a_res_url).payload
        project_a = Project(a_res)
        project_d = project_a.order['| project B | project D']
        project_d.rename('Renamed Project D')
        # This works because .rename() implementation calls shoji Entity.edit
        # which will make the request an update the resource's internal payload
        # as well. If this passes it means that Scrunch is correct and pycrunch
        # did its thing.
        self.assertEqual(project_d.resource.body.name, 'Renamed Project D')
        self.assertEqual(project_d.name, 'Renamed Project D')

    def test_move_things(self):
        a_res_url = 'http://example.com/api/projects/A/'
        dataset_url = 'http://example.com/api/datasets/1/'
        session = self.make_tree()
        project_a = Project(session.get(a_res_url).payload)
        project_c = project_a.order['| project C ']
        project_d = project_a.order['| project B | project D']
        dataset = Mock(url=dataset_url)

        # Moving project C under project D
        project_d.move_here([project_c, dataset])

        # After a move_here there is a PATCH and a GET and a GET for each item
        # the PATCH performs the changes and the GET is a resource.refresh()
        # The last request was made on the `dataset` variable which is a Mock
        # so doesn't register a request
        dataset.resource.refresh.assert_called_once()
        patch_request = session.requests[-3]
        refresh_request = session.requests[-2]
        self.assertEqual(refresh_request.method, 'GET')
        self.assertEqual(refresh_request.url, project_d.url)
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(patch_request.url, project_d.url)
        self.assertEqual(json.loads(patch_request.body), {
            'element': 'shoji:entity',
            'body': {},
            'index': {
                project_c.url: {},
                dataset.url: {}
            },
            'graph': [project_c.url, dataset.url]
        })

    def test_move_project(self):
        catalog_url = 'http://example.com/api/projects/'
        a_res_url = 'http://example.com/api/projects/A/'
        d_res_url = 'http://example.com/api/projects/D/'
        session = self.make_tree()
        project_a = Project(session.get(a_res_url).payload)
        project_d = Project(session.get(d_res_url).payload)

        with self.assertRaises(InvalidPathError):
            # Not a root path, invalid.
            project_d.move("%s" % project_a.name)

        # Moving to A
        project_d.move("| %s" % project_a.name)

        # Will have to iteratively nagivate the path making requests to
        # projects root and then to A
        request1 = session.requests[-5]
        request2 = session.requests[-4]
        self.assertEqual(request1.url, catalog_url)
        self.assertEqual(request2.url, a_res_url)

        patch_request = session.requests[-3]
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(patch_request.url, project_a.url)
        index = json.loads(patch_request.body)['index']
        self.assertEqual(index, {
            project_d.url: {},
        })

    def test_place(self):
        a_res_url = 'http://example.com/api/projects/A/'
        dataset1_url = 'http://example.com/api/datasets/1/'
        dataset2_url = 'http://example.com/api/datasets/2/'
        session = self.make_tree()
        project_a = Project(session.get(a_res_url).payload)
        project_b = project_a.order['| project B']
        project_d = project_a.order['| project B | project D']
        dataset1 = Mock(url=dataset1_url)
        dataset2 = Mock(url=dataset2_url)
        dataset1.name = 'Dataset 1'
        dataset2.name = 'Dataset 2'

        # Do a .place call
        project_a.place(dataset1, '| project B', before='project D')

        # After a move_here there is a PATCH and a GET
        # the PATCH performs the changes and the GET is a resource.refresh()
        patch_request = session.requests[-2]
        self.assertEqual(patch_request.method, 'PATCH')

        # Note the patch is to project B even though we did `.place` on
        # project A, but the target path pointed to B
        self.assertEqual(patch_request.url, project_b.url)
        self.assertEqual(json.loads(patch_request.body), {
            'element': 'shoji:entity',
            'body': {},
            'index': {
                dataset1.url: {}
            },
            # Note how the graph sent includes dataset1.url before project D
            'graph': [dataset1.url, project_d.url]
        })

        # Since the PATCH did not really update the server or the session
        # test fixtures, we need to update the fixtures to reflect the fact
        # that they've been modified by the recent PATCH request
        session.adapter.fixtures[project_b.url]['index'][dataset1.url] = {
            'name': dataset1.name,
            'type': 'dataset'
        }
        session.adapter.fixtures[project_b.url]['graph'] = [dataset1.url, project_d.url]
        project_a.place(dataset2, '| project B', after='Dataset 1')
        patch_request = session.requests[-2]
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(patch_request.url, project_b.url)
        self.assertEqual(json.loads(patch_request.body), {
            'element': 'shoji:entity',
            'body': {},
            'index': {
                dataset2.url: {}
            },
            # Dataset 2 got placed after dataset 1 :)
            'graph': [dataset1.url, dataset2.url, project_d.url]
        })

    def test_reorder(self):
        session = self.make_tree()
        a_res_url = 'http://example.com/api/projects/A/'
        b_res_url = 'http://example.com/api/projects/B/'
        c_res_url = 'http://example.com/api/projects/C/'
        project_a = Project(session.get(a_res_url).payload)
        project_a.reorder(["project C", "project B"])
        patch_request = session.requests[-2]
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(patch_request.url, project_a.url)
        self.assertEqual(json.loads(patch_request.body), {
            'element': 'shoji:entity',
            'body': {},
            'index': {},
            'graph': [c_res_url, b_res_url]
        })

    def test_move(self):
        session = self.make_tree()
        a_res_url = 'http://example.com/api/projects/A/'
        b_res_url = 'http://example.com/api/projects/B/'
        c_res_url = 'http://example.com/api/projects/C/'
        project_a = Project(session.get(a_res_url).payload)
        project_a.reorder(["project C", "project B"])
        patch_request = session.requests[-2]
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(patch_request.url, project_a.url)
        self.assertEqual(json.loads(patch_request.body), {
            'element': 'shoji:entity',
            'body': {},
            'index': {},
            'graph': [c_res_url, b_res_url]
        })

    def test_is_root(self):
        a_res_url = 'http://example.com/api/projects/A/'
        session = self.make_tree()
        project_a = Project(session.get(a_res_url).payload)
        project_b = project_a.order['| project B']
        self.assertTrue(project_a.is_root)
        self.assertFalse(project_b.is_root)

    def test_children(self):
        session = self.make_tree()
        a_res_url = 'http://example.com/api/projects/A/'
        project_a = Project(session.get(a_res_url).payload)
        # Get instantiated correctly
        self.assertTrue(all(isinstance(c, Project) for c in project_a.children))
        # Get iterated on the right order
        self.assertEqual([c.url for c in project_a.children],
                         project_a.resource.graph)
        # First request before .child should .refresh() the resource to
        # ensure fresh data
        refresh_request = session.requests[0]
        self.assertEqual(refresh_request.method, 'GET')
        self.assertEqual(refresh_request.url, a_res_url)

    def test_delete_project(self):
        mock_resource = Mock()
        project = Project(mock_resource)
        project.delete()
        mock_resource.delete.assert_called_once()


class TestPersonalProject(TestCase):
    PERSONAL_URL = "http://example.com/api/projects/personal/"

    def _get_root(self):
        session = MockSession()
        root_url = "http://example.com/api/"
        projects_url = "http://example.com/api/projects/"
        session.add_fixture(root_url, {
            "self": root_url,
            "element": "shoji:catalog",
            "catalogs": {
                "projects": projects_url
            },
            "index": {}
        })
        session.add_fixture(projects_url, {
            "self": projects_url,
            "element": "shoji:catalog",
            "catalogs": {
                "personal": self.PERSONAL_URL
            },
            "index": {}
        })
        session.add_fixture(self.PERSONAL_URL, {
            "self": self.PERSONAL_URL,
            "element": "shoji:catalog",
            "index": {}
        })
        return session.get(root_url).payload

    def test_get_personal(self):
        root = self._get_root()
        personal = get_personal_project(root)
        self.assertEqual(personal.url, self.PERSONAL_URL)

    def test_move_to_personal(self):
        root = self._get_root()
        a_res_url = 'http://example.com/api/projects/A/'
        root.session.add_fixture(a_res_url, {
            'self': a_res_url,
            'element': 'shoji:entity',
            'catalogs': {
                'project': 'http://example.com/api/projects/'
            },
            'body': {
                'name': 'project A'
            },
            'index': {},
            'graph': []
        })
        personal = get_personal_project(root)
        project_a = Project(root.session.get(a_res_url).payload)
        personal.move_here(project_a)
        patch_request = root.session.requests[-3]
        self_refresh_request = root.session.requests[-2]
        child_refresh_request = root.session.requests[-1]
        self.assertEqual(child_refresh_request.method, 'GET')
        self.assertEqual(self_refresh_request.method, 'GET')
        self.assertEqual(self_refresh_request.url, self.PERSONAL_URL)
        self.assertEqual(patch_request.url, self.PERSONAL_URL)
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(json.loads(patch_request.body)['index'], {
            project_a.url: {}
        })



# coding: utf-8

from scrunch.helpers import subvar_alias

# These are the categories that multiple response use. Selected and Not Selected
MR_CATS = [
    {"id": 1, "name": "Selected", "missing": False, "numeric_value": None, "selected": True},
    {"id": 2, "name": "Not selected", "missing": False, "numeric_value": None, "selected": False}
]

NEWS_DATASET = {
    "caseid": {
        "name": "Case ID",
        "type": "numeric"
    },
    "age": {
        "name": "Age",
        "type": 'numeric',
    },
    "newssource": {
        "name": "News source",
        "type": "multiple_response",
        "categories": MR_CATS,
        "subreferences": [{
            "name": "Facebook",
            "alias": "newssource_1"
        }, {
            "name": "Twitter",
            "alias": "newssource_2"
        }, {
            "name": "Google news",
            "alias": "newssource_3"
        }, {
            "name": "Reddit",
            "alias": "newssource_4"
        }, {
            "name": "NY Times (Print)",
            "alias": "newssource_5"
        }, {
            "name": "Washington Post (Print)",
            "alias": "newssource_6"
        }, {
            "name": "NBC News",
            "alias": "newssource_7"
        }, {
            "name": "NPR",
            "alias": "newssource_8"
        }, {
            "name": "Fox",
            "alias": "newssource_9"
        }, {
            "name": "Local radio",
            "alias": "newssource_10"
        }]
    },
    "socialmedia": {
        "name": "Accounts in social media",
        "type": "multiple_response",
        "categories": MR_CATS,
        "subreferences": [{
            "name": "Facebook",
            "alias": "socialmedia_1"
        }, {
            "name": "Twitter",
            "alias": "socialmedia_2"
        }, {
            "name": "Google+",
            "alias": "socialmedia_3"
        }, {
            "name": "VK",
            "alias": "socialmedia_4"
        }]
    },
    "gender": {
        "name": "Gender",
        "type": "categorical",
        "categories": [
            {"id": 1, "name": "Female", "missing": False, "numeric_value": None},
            {"id": 2, "name": "Male", "missing": False, "numeric_value": None},
            {"id": -1, "name": "No Data", "missing": True, "numeric_value": None},
        ]
    }
}

RECODES_CSV_OUTPUT = """newssource_1,newssource_2,newssource_3,newssource_4,newssource_5,newssource_6,newssource_7,newssource_8,newssource_9,newssource_10,age,socialmedia_1,socialmedia_2,socialmedia_3,socialmedia_4,caseid,gender,agerange,origintype_1,origintype_2,origintype_3,origintype_4,origintype_copy_1,origintype_copy_2,origintype_copy_3,origintype_copy_4,onlinenewssource_1,onlinenewssource_2,over35
1,2,1,1,2,2,2,2,2,2,25.0,1,1,2,2,1.0,1,3,1,2,2,2,1,2,2,2,1,2,1
1,2,2,2,1,2,1,1,2,1,41.0,2,2,2,2,2.0,2,4,1,1,1,1,1,1,1,1,1,1,2
1,1,1,1,2,1,1,1,2,2,33.0,1,1,1,1,3.0,2,3,1,1,1,1,1,1,1,1,1,1,1
1,2,1,1,1,2,1,2,2,2,38.0,1,1,2,2,4.0,1,4,1,1,1,2,1,1,1,2,1,1,2
2,2,2,2,1,1,2,2,1,1,50.0,2,2,2,2,5.0,1,4,2,1,1,1,2,1,1,1,2,1,2
2,2,2,2,2,2,2,2,2,1,61.0,2,2,2,2,7.0,2,5,2,2,2,1,2,2,2,1,2,1,2"""


NEWS_DATASET_ROWS = {
    "caseid": [1, 2, 3, 4, 5, 6, 7],
    "age": [25, 41, 33, 38, 50, 17, 61],
    "newssource_1": [1, 1, 1, 1, 2, 1, 2],
    "newssource_2": [2, 2, 1, 2, 2, 1, 2],
    "newssource_3": [1, 2, 1, 1, 2, 1, 2],
    "newssource_4": [1, 2, 1, 1, 2, 2, 2],
    "newssource_5": [2, 1, 2, 1, 1, 2, 2],
    "newssource_6": [2, 2, 1, 2, 1, 2, 2],
    "newssource_7": [2, 1, 1, 1, 2, 2, 2],
    "newssource_8": [2, 1, 1, 2, 2, 2, 2],
    "newssource_9": [2, 2, 2, 2, 1, 2, 2],
    "newssource_10": [2, 1, 2, 2, 1, 2, 1],
    "socialmedia_1": [1, 2, 1, 1, 2, 1, 2],
    "socialmedia_2": [1, 2, 1, 1, 2, 1, 2],
    "socialmedia_3": [2, 2, 1, 2, 2, 1, 2],
    "socialmedia_4": [2, 2, 1, 2, 2, 2, 2],
    "gender": [1, 2, 2, 1, 1, 1, 2]
}


def mr_in(ds, mr_alias, subvars):
    """
    Temporary helper until scrunch can parse correctly the expression:
     mr.has_any([sv1, sv2...])
    """
    variables = ds.resource.variables.by('alias')
    mr = variables[mr_alias].entity
    subvariables = mr.subvariables.by('alias')
    return {
        'function': 'any',
        'args': [{
            'variable': mr.self
        }, {
            'column': [subvariables[subvar_alias(mr_alias, sv)].id for sv in subvars]
        }]
    }
